@@ -15,19 +15,30 @@ import (
 var cycle *color.Cycle = &color.Cycle{}
 
 type Engine struct {
-	Config *config.Config
-	Env    runtime.Environment
-	Plain  bool
-
-	prompt            strings.Builder
-	currentLineLength int
-	rprompt           string
-	rpromptLength     int
-
+	Env                   runtime.Environment
+	Config                *config.Config
 	activeSegment         *config.Segment
 	previousActiveSegment *config.Segment
+	rprompt               string
+	Overflow              config.Overflow
+	prompt                strings.Builder
+	currentLineLength     int
+	rpromptLength         int
+	Padding               int
+	Plain                 bool
 }
 
+const (
+	PRIMARY   = "primary"
+	TRANSIENT = "transient"
+	DEBUG     = "debug"
+	SECONDARY = "secondary"
+	RIGHT     = "right"
+	TOOLTIP   = "tooltip"
+	VALID     = "valid"
+	ERROR     = "error"
+)
+
 func (e *Engine) write(text string) {
 	e.prompt.WriteString(text)
 }
@@ -65,12 +76,17 @@ func (e *Engine) canWriteRightBlock(length int, rprompt bool) (int, bool) {
 
 	canWrite := availableSpace >= promptBreathingRoom
 
+	// reset the available space when we can't write so we can fill the line
+	if !canWrite {
+		availableSpace = consoleWidth - length
+	}
+
 	return availableSpace, canWrite
 }
 
 func (e *Engine) pwd() {
 	// only print when relevant
-	if len(e.Config.PWD) == 0 && !e.Config.OSC99 {
+	if len(e.Config.PWD) == 0 {
 		return
 	}
 
@@ -85,16 +101,9 @@ func (e *Engine) pwd() {
 		pwd = strings.ReplaceAll(pwd, `\`, `/`)
 	}
 
-	// Backwards compatibility for deprecated OSC99
-	if e.Config.OSC99 {
-		e.write(terminal.Pwd(terminal.OSC99, "", "", pwd))
-		return
-	}
-
 	// Allow template logic to define when to enable the PWD (when supported)
 	tmpl := &template.Text{
 		Template: e.Config.PWD,
-		Env:      e.Env,
 	}
 
 	pwdType, err := tmpl.Render()
@@ -108,20 +117,25 @@ func (e *Engine) pwd() {
 }
 
 func (e *Engine) getNewline() string {
-	// WARP terminal will remove \n from the prompt, so we hack a newline in.
-	// For Elvish, we do this to prevent cutting off a right-aligned block.
-	if e.isWarp() || e.Env.Shell() == shell.ELVISH {
+	newline := "\n"
+
+	if e.Plain || e.Env.Flags().Debug {
+		return newline
+	}
+
+	// Warp terminal will remove a newline character ('\n') from the prompt, so we hack it in.
+	// For Elvish on Windows, we do this to prevent cutting off a right-aligned block.
+	// For Tcsh, we do this to prevent a newline character from being printed.
+	switch {
+	case e.isWarp():
+		fallthrough
+	case e.Env.Shell() == shell.ELVISH && e.Env.GOOS() == runtime.WINDOWS:
+		fallthrough
+	case e.Env.Shell() == shell.TCSH:
 		return terminal.LineBreak()
+	default:
+		return newline
 	}
-
-	// TCSH needs a space before the LITERAL newline character or it will not render correctly
-	// don't ask why, it be like that sometimes.
-	// https://unix.stackexchange.com/questions/99101/properly-defining-a-multi-line-prompt-in-tcsh#comment1342462_322189
-	if e.Env.Shell() == shell.TCSH {
-		return ` \n`
-	}
-
-	return "\n"
 }
 
 func (e *Engine) writeNewline() {
@@ -145,11 +159,18 @@ func (e *Engine) shouldFill(filler string, padLength int) (string, bool) {
 		return "", false
 	}
 
-	if padLength <= 0 {
+	tmpl := &template.Text{
+		Template: filler,
+		Context:  e,
+	}
+
+	var err error
+	if filler, err = tmpl.Render(); err != nil {
 		return "", false
 	}
 
 	// allow for easy color overrides and templates
+	terminal.SetColors("default", "default")
 	terminal.Write("", "", filler)
 	filler, lenFiller := terminal.String()
 	if lenFiller == 0 {
@@ -165,7 +186,6 @@ func (e *Engine) shouldFill(filler string, padLength int) (string, bool) {
 func (e *Engine) getTitleTemplateText() string {
 	tmpl := &template.Text{
 		Template: e.Config.ConsoleTitleTemplate,
-		Env:      e.Env,
 	}
 	if text, err := tmpl.Render(); err == nil {
 		return text
@@ -174,27 +194,15 @@ func (e *Engine) getTitleTemplateText() string {
 }
 
 func (e *Engine) renderBlock(block *config.Block, cancelNewline bool) bool {
+	text, length := e.writeBlockSegments(block)
+
+	// do not print anything when we don't have any text
+	if length == 0 {
+		return false
+	}
+
 	defer e.applyPowerShellBleedPatch()
 
-	// This is deprecated but we leave it in to not break configs
-	// It is encouraged to use "newline": true on block level
-	// rather than the standalone linebreak block
-	if block.Type == config.LineBreak {
-		// do not print a newline to avoid a leading space
-		// when we're printing the first primary prompt in
-		// the shell
-		if !cancelNewline {
-			e.writeNewline()
-		}
-		return false
-	}
-
-	block.Init(e.Env)
-
-	if !block.Enabled() {
-		return false
-	}
-
 	// do not print a newline to avoid a leading space
 	// when we're printing the first primary prompt in
 	// the shell
@@ -202,19 +210,8 @@ func (e *Engine) renderBlock(block *config.Block, cancelNewline bool) bool {
 		e.writeNewline()
 	}
 
-	text, length := e.renderBlockSegments(block)
-
-	// do not print anything when we don't have any text
-	if length == 0 {
-		return false
-	}
-
-	switch block.Type { //nolint:exhaustive
+	switch block.Type {
 	case config.Prompt:
-		if block.VerticalOffset != 0 {
-			e.write(terminal.ChangeLine(block.VerticalOffset))
-		}
-
 		if block.Alignment == config.Left {
 			e.currentLineLength += length
 			e.write(text)
@@ -229,6 +226,7 @@ func (e *Engine) renderBlock(block *config.Block, cancelNewline bool) bool {
 
 		// we can't print the right block as there's not enough room available
 		if !OK {
+			e.Overflow = block.Overflow
 			switch block.Overflow {
 			case config.Break:
 				e.writeNewline()
@@ -245,6 +243,7 @@ func (e *Engine) renderBlock(block *config.Block, cancelNewline bool) bool {
 
 		defer func() {
 			e.currentLineLength = 0
+			e.Overflow = ""
 		}()
 
 		// validate if we have a filler and fill if needed
@@ -287,50 +286,6 @@ func (e *Engine) applyPowerShellBleedPatch() {
 	e.write(terminal.ClearAfter())
 }
 
-func (e *Engine) renderBlockSegments(block *config.Block) (string, int) {
-	e.filterSegments(block)
-
-	for i, segment := range block.Segments {
-		if colors, newCycle := cycle.Loop(); colors != nil {
-			cycle = &newCycle
-			segment.Foreground = colors.Foreground
-			segment.Background = colors.Background
-		}
-
-		if i == 0 && len(block.LeadingDiamond) > 0 {
-			segment.LeadingDiamond = block.LeadingDiamond
-		}
-
-		if i == len(block.Segments)-1 && len(block.TrailingDiamond) > 0 {
-			segment.TrailingDiamond = block.TrailingDiamond
-		}
-
-		e.setActiveSegment(segment)
-		e.renderActiveSegment()
-	}
-
-	e.writeSeparator(true)
-
-	e.activeSegment = nil
-	e.previousActiveSegment = nil
-
-	return terminal.String()
-}
-
-func (e *Engine) filterSegments(block *config.Block) {
-	segments := make([]*config.Segment, 0)
-
-	for _, segment := range block.Segments {
-		if !segment.Enabled && segment.ResolveStyle() != config.Accordion {
-			continue
-		}
-
-		segments = append(segments, segment)
-	}
-
-	block.Segments = segments
-}
-
 func (e *Engine) setActiveSegment(segment *config.Segment) {
 	e.activeSegment = segment
 	terminal.Interactive = segment.Interactive
@@ -342,7 +297,7 @@ func (e *Engine) renderActiveSegment() {
 
 	switch e.activeSegment.ResolveStyle() {
 	case config.Plain, config.Powerline:
-		terminal.Write(color.Background, color.Foreground, e.activeSegment.Text)
+		terminal.Write(color.Background, color.Foreground, e.activeSegment.Text())
 	case config.Diamond:
 		background := color.Transparent
 
@@ -351,10 +306,10 @@ func (e *Engine) renderActiveSegment() {
 		}
 
 		terminal.Write(background, color.Background, e.activeSegment.LeadingDiamond)
-		terminal.Write(color.Background, color.Foreground, e.activeSegment.Text)
+		terminal.Write(color.Background, color.Foreground, e.activeSegment.Text())
 	case config.Accordion:
 		if e.activeSegment.Enabled {
-			terminal.Write(color.Background, color.Foreground, e.activeSegment.Text)
+			terminal.Write(color.Background, color.Foreground, e.activeSegment.Text())
 		}
 	}
 
@@ -364,6 +319,10 @@ func (e *Engine) renderActiveSegment() {
 }
 
 func (e *Engine) writeSeparator(final bool) {
+	if e.activeSegment == nil {
+		return
+	}
+
 	isCurrentDiamond := e.activeSegment.ResolveStyle() == config.Diamond
 	if final && isCurrentDiamond {
 		terminal.Write(color.Transparent, color.Background, e.activeSegment.TrailingDiamond)
@@ -506,23 +465,38 @@ func (e *Engine) adjustTrailingDiamondColorOverrides() {
 	}
 }
 
+func (e *Engine) rectifyTerminalWidth(diff int) {
+	// Since the terminal width may not be given by the CLI flag, we should always call this here.
+	_, err := e.Env.TerminalWidth()
+	if err != nil {
+		// Skip when we're unable to determine the terminal width.
+		return
+	}
+
+	e.Env.Flags().TerminalWidth += diff
+}
+
 // New returns a prompt engine initialized with the
 // given configuration options, and is ready to print any
 // of the prompt components.
 func New(flags *runtime.Flags) *Engine {
-	env := &runtime.Terminal{
-		CmdFlags: flags,
-	}
+	flags.Config = config.Path(flags.Config)
+	cfg := config.Load(flags.Config, flags.Shell, flags.Migrate)
 
-	env.Init()
-	cfg := config.Load(env)
+	env := &runtime.Terminal{}
+	env.Init(flags)
 
-	env.Var = cfg.Var
-	flags.HasTransient = cfg.TransientPrompt != nil
+	template.Init(env, cfg.Var)
+
+	flags.HasExtra = cfg.DebugPrompt != nil ||
+		cfg.SecondaryPrompt != nil ||
+		cfg.TransientPrompt != nil ||
+		cfg.ValidLine != nil ||
+		cfg.ErrorLine != nil
 
 	terminal.Init(env.Shell())
-	terminal.BackgroundColor = cfg.TerminalBackground.ResolveTemplate(env)
-	terminal.Colors = cfg.MakeColors()
+	terminal.BackgroundColor = cfg.TerminalBackground.ResolveTemplate()
+	terminal.Colors = cfg.MakeColors(env)
 	terminal.Plain = flags.Plain
 
 	eng := &Engine{
@@ -532,9 +506,16 @@ func New(flags *runtime.Flags) *Engine {
 	}
 
 	switch env.Shell() {
-	case shell.ELVISH:
-		// In Elvish, a prompt line will always wrap when the cursor reaches the rightmost cell on the terminal screen.
-		// We have to reduce the terminal width by 1 so a right-aligned block will not be broken.
+	case shell.XONSH:
+		// In Xonsh, the behavior of wrapping at the end of a prompt line is inconsistent across platforms.
+		// On Windows, it wraps before the rightmost cell on the terminal screen, that is, the rightmost cell is never available for a prompt line.
+		if eng.Env.GOOS() == runtime.WINDOWS {
+			eng.rectifyTerminalWidth(-1)
+		}
+	case shell.TCSH, shell.ELVISH:
+		// In Tcsh, newlines in a prompt are badly translated.
+		// No silver bullet here. We have to reduce the terminal width by 1 so a right-aligned block will not be broken.
+		// In Elvish, the behavior is similar to that in Xonsh, but we do this for all platforms.
 		eng.rectifyTerminalWidth(-1)
 	case shell.PWSH, shell.PWSH5:
 		// when in PowerShell, and force patching the bleed bug
@@ -548,14 +529,3 @@ func New(flags *runtime.Flags) *Engine {
 
 	return eng
 }
-
-func (e *Engine) rectifyTerminalWidth(diff int) {
-	// Since the terminal width may not be given by the CLI flag, we should always call this here.
-	_, err := e.Env.TerminalWidth()
-	if err != nil {
-		// Skip when we're unable to determine the terminal width.
-		return
-	}
-
-	e.Env.Flags().TerminalWidth += diff
-}
